import{p as e,f as o}from"./flowDb-956e92f1-a4991264.js";import{f as t,g as a}from"./styles-c10674c1-c1d8f7e9.js";import{u as i}from"./index-7794b245.js";import"./graph-be6ef5d8.js";import"./layout-d649bc98.js";import"./index-3862675e-d9ce8994.js";import"./clone-21f8a43d.js";import"./edges-e0da2a9e-df20501d.js";import"./createText-2e5e7dd3-75dfaa67.js";import"./line-95add810.js";import"./array-9f3ba611.js";import"./path-53f90ab3.js";import"./channel-e265ea59.js";const M={parser:e,db:o,renderer:t,styles:a,init:r=>{r.flowchart||(r.flowchart={}),r.flowchart.arrowMarkerAbsolute=r.arrowMarkerAbsolute,i({flowchart:{arrowMarkerAbsolute:r.arrowMarkerAbsolute}}),t.setConf(r.flowchart),o.clear(),o.setGen("gen-2")}};export{M as diagram};
