DEFAULT_ENCODING = "utf-8"

READ_YAML_FILE_CACHE_MAXSIZE_ENV_VAR = "READ_YAML_FILE_CACHE_MAXSIZE"
DEFAULT_READ_YAML_FILE_CACHE_MAXSIZE = 256
RASA_PRO_BETA_PREDICATES_IN_RESPONSE_CONDITIONS_ENV_VAR_NAME = (
    "RASA_PRO_BETA_PREDICATES_IN_RESPONSE_CONDITIONS"
)
